"""RAG Retrieval package."""

from __future__ import annotations

from rago.retrieval.base import RetrievalBase, StringRet
from rago.retrieval.file import PDFPathRet

__all__ = [
    'PDFPathRet',
    'RetrievalBase',
    'StringRet',
]



"""Provide base interfaces."""

from __future__ import annotations

from abc import ABC
from typing import Any, Optional

from rago.extensions.cache import Cache


class RagoBase(ABC):
    """Define base interface for RAG step classes."""

    api_key: str = ''
    cache: Optional[Cache] = None
    logs: dict[str, Any] = {}  # noqa: RUF012

    def __init__(
        self,
        api_key: str = '',
        cache: Optional[Cache] = None,
        logs: dict[str, Any] = {},
    ) -> None:
        self.api_key = api_key
        self.cache = cache
        self.logs = logs

    def _get_cache(self, key: Any) -> Any:
        if not self.cache:
            return
        return self.cache.load(key)

    def _save_cache(self, key: Any, data: Any) -> None:
        if not self.cache:
            return
        self.cache.save(key, data)



"""Base classes for database."""

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Iterable

from typeguard import typechecked


@typechecked
class DBBase:
    """Base class for vector database."""

    index: Any

    @abstractmethod
    def embed(self, documents: Any) -> None:
        """Embed the documents into the database."""
        ...

    @abstractmethod
    def search(
        self, query_encoded: Any, top_k: int = 2
    ) -> tuple[Iterable[float], Iterable[int]]:
        """Search a query from documents."""
        ...



"""The base classes for text splitter."""

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Iterable


class TextSplitterBase:
    """The base text splitter class."""

    chunk_size: int = 500
    chunk_overlap: int = 100
    splitter_name: str = ''
    splitter: Any = None

    # defaults

    default_chunk_size: int = 500
    default_chunk_overlap: int = 100
    default_splitter_name: str = ''
    default_splitter: Any = None

    def __init__(
        self,
        splitter_name: str = '',
        chunk_size: int = 500,
        chunk_overlap: int = 100,
    ) -> None:
        """Initialize the text splitter class."""
        self.chunk_size = chunk_size or self.default_chunk_size
        self.chunk_overlap = chunk_overlap or self.default_chunk_overlap
        self.splitter_name = splitter_name or self.default_splitter_name

        self._validate()
        self._setup()

    def _validate(self) -> None:
        """Validate if the initial parameters are valid."""
        return

    def _setup(self) -> None:
        """Set up the object according to the given parameters."""
        return

    @abstractmethod
    def split(self, text: str) -> Iterable[str]:
        """Split a text into chunks."""
        return []



"""Support langchain text splitter."""

from __future__ import annotations

from typing import List, cast

from langchain.text_splitter import RecursiveCharacterTextSplitter

from rago.retrieval.text_splitter.base import TextSplitterBase


class LangChainTextSplitter(TextSplitterBase):
    """LangChain Text Splitter class."""

    default_splitter_name: str = 'RecursiveCharacterTextSplitter'

    def _validate(self) -> None:
        """Validate if the initial parameters are valid."""
        valid_splitter_names = ['RecursiveCharacterTextSplitter']

        if self.splitter_name not in valid_splitter_names:
            raise Exception(
                f'The given splitter_name {self.splitter_name} is not valid. '
                f'Valid options are {valid_splitter_names}'
            )

    def _setup(self) -> None:
        """Set up the object according to the given parameters."""
        if self.splitter_name == 'RecursiveCharacterTextSplitter':
            self.splitter = RecursiveCharacterTextSplitter

    def split(self, text: str) -> list[str]:
        """Split text into smaller chunks for processing."""
        text_splitter = self.splitter(
            chunk_size=self.chunk_size,
            chunk_overlap=self.chunk_overlap,
            length_function=len,
            is_separator_regex=True,
        )
        return cast(List[str], text_splitter.split_text(text))



"""Package for classes about text splitter."""

from rago.retrieval.text_splitter.base import TextSplitterBase
from rago.retrieval.text_splitter.langchain import LangChainTextSplitter

__all__ = [
    'LangChainTextSplitter',
    'TextSplitterBase',
]



"""Rago."""

from importlib import metadata as importlib_metadata

from rago.core import Rago


def get_version() -> str:
    """Return the program version."""
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return '0.11.1'  # semantic-release


version = get_version()

__version__ = version
__author__ = 'Ivan Ogasawara'
__email__ = 'ivan.ogasawara@gmail.com'

__all__ = [
    'Rago',
    '__author__',
    '__email__',
    '__version__',
]



"""Base classes for generation."""

from __future__ import annotations

from abc import abstractmethod
from copy import deepcopy
from typing import Any, Optional, Type

import torch

from pydantic import BaseModel
from typeguard import typechecked

from rago.base import RagoBase
from rago.extensions.cache import Cache

DEFAULT_LOGS: dict[str, Any] = {}
DEFAULT_API_PARAMS: dict[str, Any] = {}


@typechecked
class GenerationBase(RagoBase):
    """Generic Generation class."""

    device_name: str = 'cpu'
    device: torch.device
    model: Any
    model_name: str = ''
    tokenizer: Any
    temperature: float = 0.5
    output_max_length: int = 500
    prompt_template: str = (
        'question: \n```\n{query}\n```\ncontext: ```\n{context}\n```'
    )
    structured_output: Optional[Type[BaseModel]] = None
    api_params: dict[str, Any] = {}  # noqa: RUF012

    # default parameters that can be overwritten by the derived class
    default_device_name: str = 'cpu'
    default_model_name: str = ''
    default_temperature: float = 0.5
    default_output_max_length: int = 500
    default_prompt_template: str = (
        'question: \n```\n{query}\n```\ncontext: ```\n{context}\n```'
    )
    default_api_params: dict[str, Any] = {}  # noqa: RUF012

    def __init__(
        self,
        model_name: Optional[str] = None,
        temperature: Optional[float] = None,
        prompt_template: str = '',
        output_max_length: int = 500,
        device: str = 'auto',
        structured_output: Optional[Type[BaseModel]] = None,
        api_params: dict[str, Any] = DEFAULT_API_PARAMS,
        api_key: str = '',
        cache: Optional[Cache] = None,
        logs: dict[str, Any] = DEFAULT_LOGS,
    ) -> None:
        """Initialize Generation class."""
        if logs is DEFAULT_LOGS:
            logs = {}
        super().__init__(api_key=api_key, cache=cache, logs=logs)

        self.model_name: str = (
            model_name if model_name is not None else self.default_model_name
        )
        self.output_max_length: int = (
            output_max_length or self.default_output_max_length
        )
        self.temperature: float = (
            temperature
            if temperature is not None
            else self.default_temperature
        )

        self.prompt_template: str = (
            prompt_template or self.default_prompt_template
        )
        self.structured_output: Optional[Type[BaseModel]] = structured_output
        if api_params is DEFAULT_API_PARAMS:
            api_params = deepcopy(self.default_api_params or {})

        self.api_params = api_params

        if device not in ['cpu', 'cuda', 'auto']:
            raise Exception(
                f'Device {device} not supported. Options: cpu, cuda, auto.'
            )

        cuda_available = torch.cuda.is_available()
        self.device_name: str = (
            'cpu' if device == 'cpu' or not cuda_available else 'cuda'
        )
        self.device = torch.device(self.device_name)

        self._validate()
        self._setup()

    def _validate(self) -> None:
        """Raise an error if the initial parameters are not valid."""
        return

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        return

    @abstractmethod
    def generate(
        self,
        query: str,
        context: list[str],
    ) -> str | BaseModel:
        """Generate text with optional language parameter.

        Parameters
        ----------
        query : str
            The input query or prompt.
        context : list[str]
            Additional context information for the generation.

        Returns
        -------
        str
            Generated text based on query and context.
        """
        ...



"""Base classes for the augmented step."""

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Optional, Union

import numpy as np
import numpy.typing as npt

from torch import Tensor
from typeguard import typechecked
from typing_extensions import TypeAlias

from rago.augmented.db import DBBase, FaissDB
from rago.base import RagoBase
from rago.extensions.cache import Cache

EmbeddingType: TypeAlias = Union[
    npt.NDArray[np.float64],
    Tensor,
    list[Tensor],
]

DEFAULT_LOGS: dict[str, Any] = {}


@typechecked
class AugmentedBase(RagoBase):
    """Define the base structure for Augmented classes."""

    model: Optional[Any]
    model_name: str = ''
    db: Any
    top_k: int = 0

    # default values to be overwritten by the derived classes
    default_model_name: str = ''
    default_top_k: int = 5

    def __init__(
        self,
        model_name: Optional[str] = None,
        db: DBBase = FaissDB(),
        top_k: Optional[int] = None,
        api_key: str = '',
        cache: Optional[Cache] = None,
        logs: dict[str, Any] = DEFAULT_LOGS,
    ) -> None:
        """Initialize AugmentedBase."""
        if logs is DEFAULT_LOGS:
            logs = {}
        super().__init__(api_key=api_key, cache=cache, logs=logs)

        self.db = db

        self.top_k = top_k if top_k is not None else self.default_top_k
        self.model_name = (
            model_name if model_name is not None else self.default_model_name
        )
        self.model = None

        self._validate()
        self._setup()

    def _validate(self) -> None:
        """Raise an error if the initial parameters are not valid."""
        return

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        return

    def get_embedding(self, content: list[str]) -> EmbeddingType:
        """Retrieve the embedding for a given text using OpenAI API."""
        raise Exception('Method not implemented.')

    @abstractmethod
    def search(
        self,
        query: str,
        documents: Any,
        top_k: int = 0,
    ) -> list[str]:
        """Search an encoded query into vector database."""
        ...



"""Base classes for retrieval."""

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Iterable, Optional, cast

from typeguard import typechecked

from rago.base import RagoBase
from rago.extensions.cache import Cache
from rago.retrieval.text_splitter import (
    LangChainTextSplitter,
    TextSplitterBase,
)

DEFAULT_LOGS: dict[str, Any] = {}


@typechecked
class RetrievalBase(RagoBase):
    """Base Retrieval class."""

    content: Any
    source: Any
    splitter: TextSplitterBase

    def __init__(
        self,
        source: Any,
        splitter: TextSplitterBase = LangChainTextSplitter(
            'RecursiveCharacterTextSplitter'
        ),
        api_key: str = '',
        cache: Optional[Cache] = None,
        logs: dict[str, Any] = DEFAULT_LOGS,
    ) -> None:
        """Initialize the Retrieval class."""
        if logs is DEFAULT_LOGS:
            logs = {}
        super().__init__(api_key=api_key, cache=cache, logs=logs)
        self.source = source
        self.splitter = splitter

        self._validate()
        self._setup()

    def _validate(self) -> None:
        """Validate if the source is valid, otherwise raises an exception."""
        return None

    def _setup(self) -> None:
        """Set up the object with the giving initial parameters."""
        return None

    @abstractmethod
    def get(self, query: str = '') -> Iterable[str]:
        """Get the data from the source."""
        return []


@typechecked
class StringRet(RetrievalBase):
    """
    String Retrieval class.

    This is a very generic class that assumes that the input (source) is
    already a list of strings.
    """

    def get(self, query: str = '') -> Iterable[str]:
        """Get the data from the sources."""
        return cast(list[str], self.source)



"""Classes for augmentation with OpenAI embeddings."""

from __future__ import annotations

from hashlib import sha256
from typing import cast

import numpy as np
import openai

from typeguard import typechecked

from rago.augmented.base import AugmentedBase, EmbeddingType


@typechecked
class OpenAIAug(AugmentedBase):
    """Class for augmentation with OpenAI embeddings."""

    default_model_name = 'text-embedding-3-small'
    default_top_k = 3

    def _setup(self) -> None:
        """Set up the object with initial parameters."""
        if not self.api_key:
            raise ValueError('API key for OpenAI is required.')
        openai.api_key = self.api_key
        self.model = openai.OpenAI(api_key=self.api_key)

    def get_embedding(self, content: list[str]) -> EmbeddingType:
        """Retrieve the embedding for a given text using OpenAI API."""
        cache_key = sha256(''.join(content).encode('utf-8')).hexdigest()
        cached = self._get_cache(cache_key)
        if cached is not None:
            return cast(EmbeddingType, cached)

        model = cast(openai.OpenAI, self.model)
        response = model.embeddings.create(
            input=content, model=self.model_name
        )
        result = np.array(response.data[0].embedding)
        result = result.reshape(1, result.size)

        self._save_cache(cache_key, result)

        return result

    def search(
        self, query: str, documents: list[str], top_k: int = 0
    ) -> list[str]:
        """Search an encoded query into vector database."""
        if not hasattr(self, 'db') or not self.db:
            raise Exception('Vector database (db) is not initialized.')

        # Encode the documents and query
        document_encoded = self.get_embedding(documents)
        query_encoded = self.get_embedding([query])
        top_k = top_k or self.top_k or self.default_top_k or 1

        self.db.embed(document_encoded)
        scores, indices = self.db.search(query_encoded, top_k=top_k)

        self.logs['indices'] = indices
        self.logs['scores'] = scores
        self.logs['search_params'] = {
            'query_encoded': query_encoded,
            'top_k': top_k,
        }

        retrieved_docs = [documents[i] for i in indices if i >= 0]

        return retrieved_docs



"""Llama generation module."""

from __future__ import annotations

import warnings

import torch

from langdetect import detect
from transformers import AutoModelForCausalLM, AutoTokenizer, pipeline
from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class LlamaGen(GenerationBase):
    """Llama Generation class."""

    default_model_name: str = 'meta-llama/Llama-3.2-1B'
    default_temperature: float = 0.5
    default_output_max_length: int = 500

    def _validate(self) -> None:
        """Raise an error if the initial parameters are not valid."""
        if not self.model_name.startswith('meta-llama/'):
            raise Exception(
                f'The given model name {self.model_name} is not provided '
                'by meta.'
            )

        if self.structured_output:
            warnings.warn(
                'Structured output is not supported yet in '
                f'{self.__class__.__name__}.'
            )

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_name, token=self.api_key
        )

        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_name,
            token=self.api_key,
            torch_dtype=torch.float16
            if self.device_name == 'cuda'
            else torch.float32,
        )

        self.generator = pipeline(
            'text-generation',
            model=self.model,
            tokenizer=self.tokenizer,
            device=0 if self.device_name == 'cuda' else -1,
        )

    def generate(self, query: str, context: list[str]) -> str:
        """Generate text using Llama model with language support."""
        input_text = self.prompt_template.format(
            query=query, context=' '.join(context)
        )

        # Detect and set the language code for multilingual models (optional)
        language = str(detect(query)) or 'en'
        self.tokenizer.lang_code = language

        # Generate the response with adjusted parameters

        model_params = dict(
            text_inputs=input_text,
            max_new_tokens=self.output_max_length,
            do_sample=True,
            temperature=self.temperature,
            top_k=50,  # todo: check if it is necessary
            top_p=0.95,  # todo: check if it is necessary
            num_return_sequences=1,
            eos_token_id=self.tokenizer.eos_token_id,
        )
        response = self.generator(**model_params)

        self.logs['model_params'] = model_params

        # Extract and return the answer only
        answer = str(response[0].get('generated_text', ''))
        # Strip off any redundant text after the answer itself
        return answer.split('Answer:')[-1].strip()



"""Rago is Retrieval Augmented Generation lightweight framework."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel
from typeguard import typechecked

from rago.augmented.base import AugmentedBase
from rago.generation.base import GenerationBase
from rago.retrieval.base import RetrievalBase


@typechecked
class Rago:
    """RAG class."""

    retrieval: RetrievalBase
    augmented: AugmentedBase
    generation: GenerationBase

    logs: dict[str, dict[str, Any]]

    def __init__(
        self,
        retrieval: RetrievalBase,
        augmented: AugmentedBase,
        generation: GenerationBase,
    ) -> None:
        """Initialize the RAG structure.

        Parameters
        ----------
        retrieval : RetrievalBase
            The retrieval component used to fetch relevant data based
            on the query.
        augmented : AugmentedBase
            The augmentation module responsible for enriching the
            retrieved data.
        generation : GenerationBase
            The text generation model used to generate a response based
            on the query and augmented data.
        """
        self.retrieval = retrieval
        self.augmented = augmented
        self.generation = generation
        self.logs: dict[str, dict[str, Any]] = {
            'retrieval': retrieval.logs,
            'augmented': augmented.logs,
            'generation': generation.logs,
        }

    def prompt(self, query: str, device: str = 'auto') -> str | BaseModel:
        """Run the pipeline for a specific prompt.

        Parameters
        ----------
        query : str
            The query or prompt from the user.
        device : str (default 'auto')
            Device for generation (e.g., 'auto', 'cpu', 'cuda'), by
            default 'auto'.

        Returns
        -------
        str
            Generated text based on the query and augmented data.
        """
        ret_data = self.retrieval.get(query)
        self.logs['retrieval']['result'] = ret_data

        aug_data = self.augmented.search(query, ret_data)
        self.logs['augmented']['result'] = aug_data

        gen_data = self.generation.generate(query, context=aug_data)
        self.logs['generation']['result'] = gen_data

        return gen_data



"""Classes for augmentation with hugging face."""

from __future__ import annotations

from typing import Any, cast

from sentence_transformers import SentenceTransformer
from typeguard import typechecked

from rago.augmented.base import AugmentedBase, EmbeddingType


@typechecked
class SentenceTransformerAug(AugmentedBase):
    """Class for augmentation with Hugging Face."""

    default_model_name = 'paraphrase-MiniLM-L12-v2'
    default_top_k = 3

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        self.model = SentenceTransformer(self.model_name)

    def get_embedding(self, content: list[str]) -> EmbeddingType:
        """Retrieve the embedding for a given text using OpenAI API."""
        model = cast(SentenceTransformer, self.model)
        return model.encode(content)

    def search(self, query: str, documents: Any, top_k: int = 0) -> list[str]:
        """Search an encoded query into vector database."""
        if not self.model:
            raise Exception('The model was not created.')

        document_encoded = self.get_embedding(documents)
        query_encoded = self.get_embedding([query])
        top_k = top_k or self.top_k or self.default_top_k or 1

        self.db.embed(document_encoded)

        scores, indices = self.db.search(query_encoded, top_k=top_k)

        retrieved_docs = [documents[i] for i in indices]

        self.logs['indices'] = indices
        self.logs['scores'] = scores
        self.logs['search_params'] = {
            'query_encoded': query_encoded,
            'top_k': top_k,
        }

        return retrieved_docs



"""Base classes for retrieval."""

from __future__ import annotations

from pathlib import Path
from typing import Iterable, cast

from typeguard import typechecked

from rago.retrieval.base import RetrievalBase
from rago.retrieval.tools.pdf import extract_text_from_pdf, is_pdf


@typechecked
class FilePathRet(RetrievalBase):
    """File Retrieval class."""

    def _validate(self) -> None:
        """Validate if the source is valid, otherwise raises an exception."""
        if not isinstance(self.source, (str, Path)):
            raise Exception('Argument source should be an string or a Path.')

        source_path = Path(self.source)
        if not source_path.exists():
            raise Exception("File doesn't exist.")


@typechecked
class PDFPathRet(FilePathRet):
    """PDFPathRet Retrieval class."""

    def _validate(self) -> None:
        """Validate if the source is valid, otherwise raises an exception."""
        super()._validate()
        if not is_pdf(self.source):
            raise Exception('Given file is not a PDF.')

    def get(self, query: str = '') -> Iterable[str]:
        """Get the data from the source."""
        cache_key = self.source
        cached = self._get_cache(cache_key)
        if cached is not None:
            return cast(Iterable[str], cached)

        text = extract_text_from_pdf(self.source)

        self.logs['text'] = text

        result = self.splitter.split(text)

        self._save_cache(cache_key, result)

        return result



"""Provide an extension for caching."""

from __future__ import annotations

from abc import abstractmethod
from hashlib import sha256
from pathlib import Path
from typing import Any

import joblib

from typeguard import typechecked

from rago.extensions.base import Extension


@typechecked
class Cache(Extension):
    """Define an extension base for caching."""

    @abstractmethod
    def load(self, key: Any) -> Any:
        """Load the cache for given key."""
        raise Exception(f'Load method is not implemented: {key}')

    @abstractmethod
    def save(self, key: Any, data: Any) -> None:
        """Save the cache for given key."""
        raise Exception(f'Save method is not implemented: {key}')


@typechecked
class CacheFile(Cache):
    """Define a extra step for caching."""

    target_dir: Path

    def __init__(self, target_dir: Path) -> None:
        self.target_dir = target_dir
        self.target_dir.mkdir(parents=True, exist_ok=True)

    def get_file_path(self, key: Any) -> Path:
        """Return the file path."""
        ref = sha256(str(key).encode('utf-8')).hexdigest()
        return self.target_dir / f'{ref}.pkl'

    def load(self, key: Any) -> Any:
        """Load the cache for given key."""
        file_path = self.get_file_path(key)
        if not file_path.exists():
            return
        return joblib.load(file_path)

    def save(self, key: Any, data: Any) -> None:
        """Load the cache for given key."""
        file_path = self.get_file_path(key)
        joblib.dump(data, file_path)



"""Module for faiss database."""

from __future__ import annotations

from typing import Any, Iterable

import faiss

from typeguard import typechecked

from rago.augmented.db.base import DBBase


@typechecked
class FaissDB(DBBase):
    """Faiss Database."""

    def embed(self, documents: Any) -> None:
        """Embed the documents into the database."""
        self.index = faiss.IndexFlatL2(documents.shape[1])
        self.index.add(documents)

    def search(
        self, query_encoded: Any, top_k: int = 2
    ) -> tuple[Iterable[float], Iterable[int]]:
        """Search an encoded query into vector database."""
        distances, indices = self.index.search(query_encoded, top_k)
        return distances, indices[0]



"""OpenAI Generation Model class for flexible GPT-based text generation."""

from __future__ import annotations

from typing import cast

import instructor
import openai

from pydantic import BaseModel
from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class OpenAIGen(GenerationBase):
    """OpenAI generation model for text generation."""

    default_model_name = 'gpt-3.5-turbo'
    default_api_params = {  # noqa: RUF012
        'top_p': 1.0,
        'frequency_penalty': 0.0,
        'presence_penalty': 0.0,
    }

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        model = openai.OpenAI(api_key=self.api_key)

        self.model = (
            instructor.from_openai(model) if self.structured_output else model
        )

    def generate(
        self,
        query: str,
        context: list[str],
    ) -> str | BaseModel:
        """Generate text using OpenAI's API with dynamic model support."""
        input_text = self.prompt_template.format(
            query=query, context=' '.join(context)
        )

        if not self.model:
            raise Exception('The model was not created.')

        api_params = (
            self.api_params if self.api_params else self.default_api_params
        )

        model_params = dict(
            model=self.model_name,
            messages=[{'role': 'user', 'content': input_text}],
            max_tokens=self.output_max_length,
            temperature=self.temperature,
            **api_params,
        )

        if self.structured_output:
            model_params['response_model'] = self.structured_output

        response = self.model.chat.completions.create(**model_params)

        self.logs['model_params'] = model_params

        has_choices = hasattr(response, 'choices')

        if has_choices and isinstance(response.choices, list):
            return cast(str, response.choices[0].message.content.strip())
        return cast(BaseModel, response)



"""Hugging Face classes for text generation."""

from __future__ import annotations

import warnings

import torch

from transformers import T5ForConditionalGeneration, T5Tokenizer
from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class HuggingFaceGen(GenerationBase):
    """HuggingFaceGen."""

    default_model_name = 't5-small'

    def _validate(self) -> None:
        if self.model_name != 't5-small':
            raise Exception(
                f'The given model {self.model_name} is not supported.'
            )

        if self.structured_output:
            warnings.warn(
                'Structured output is not supported yet in '
                f'{self.__class__.__name__}.'
            )

    def _setup(self) -> None:
        """Set models to t5-small models."""
        self.tokenizer = T5Tokenizer.from_pretrained(self.model_name)
        model = T5ForConditionalGeneration.from_pretrained(self.model_name)
        self.model = model.to(self.device)

    def generate(self, query: str, context: list[str]) -> str:
        """Generate the text from the query and augmented context."""
        with torch.no_grad():
            input_text = self.prompt_template.format(
                query=query, context=' '.join(context)
            )
            input_ids = self.tokenizer.encode(
                input_text,
                return_tensors='pt',
                truncation=True,
                max_length=512,
            ).to(self.device_name)

            model_params = dict(
                inputs=input_ids,
                max_length=self.output_max_length,
                pad_token_id=self.tokenizer.eos_token_id,
            )

            outputs = self.model.generate(**model_params)

            self.logs['model_params'] = model_params

            response = self.tokenizer.decode(
                outputs[0], skip_special_tokens=True
            )

        if self.device_name == 'cuda':
            torch.cuda.empty_cache()

        return str(response)



"""Augmented package."""

from __future__ import annotations

from rago.augmented.base import AugmentedBase
from rago.augmented.openai import OpenAIAug
from rago.augmented.sentence_transformer import SentenceTransformerAug
from rago.augmented.spacy import SpaCyAug

__all__ = [
    'AugmentedBase',
    'OpenAIAug',
    'SentenceTransformerAug',
    'SpaCyAug',
]



"""PDF tools."""

from __future__ import annotations

from pathlib import Path

from pypdf import PdfReader


def is_pdf(file_path: str | Path) -> bool:
    """
    Check if a file is a PDF by reading its header.

    Parameters
    ----------
    file_path : str
        Path to the file to be checked.

    Returns
    -------
    bool
        True if the file is a PDF, False otherwise.
    """
    try:
        with open(file_path, 'rb') as file:
            header = file.read(4)
            return header == b'%PDF'
    except IOError:
        return False


def extract_text_from_pdf(file_path: str) -> str:
    """
    Extract text from a PDF file using pypdf.

    The result is the same as the one returned by PyPDFLoader.
    """
    reader = PdfReader(file_path)
    pages = []
    for page in reader.pages:
        text = page.extract_text()
        if text:
            pages.append(text)
    return ' '.join(pages)



"""RAG Generation package."""

from __future__ import annotations

from rago.generation.base import GenerationBase
from rago.generation.gemini import GeminiGen
from rago.generation.hugging_face import HuggingFaceGen
from rago.generation.llama import LlamaGen
from rago.generation.openai import OpenAIGen

__all__ = [
    'GeminiGen',
    'GenerationBase',
    'HuggingFaceGen',
    'LlamaGen',
    'OpenAIGen',
]



"""GeminiGen class for text generation using Google's Gemini model."""

from __future__ import annotations

from typing import cast

import google.generativeai as genai
import instructor

from pydantic import BaseModel
from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class GeminiGen(GenerationBase):
    """Gemini generation model for text generation."""

    default_model_name: str = 'gemini-1.5-flash'

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        genai.configure(api_key=self.api_key)
        model = genai.GenerativeModel(self.model_name)

        self.model = (
            instructor.from_gemini(
                client=model,
                mode=instructor.Mode.GEMINI_JSON,
            )
            if self.structured_output
            else model
        )

    def generate(self, query: str, context: list[str]) -> str | BaseModel:
        """Generate text using Gemini model support."""
        input_text = self.prompt_template.format(
            query=query, context=' '.join(context)
        )

        if not self.structured_output:
            models_params_gen = {'contents': input_text}
            response = self.model.generate_content(**models_params_gen)
            self.logs['model_params'] = models_params_gen
            return cast(str, response.text.strip())

        messages = [
            {'role': 'user', 'content': input_text},
        ]
        model_params = {
            'messages': messages,
            'response_model': self.structured_output,
        }

        response = self.model.create(
            **model_params,
        )

        self.logs['model_params'] = model_params

        return cast(BaseModel, response)
